#include "smptestconf.h"
/*
 * Copyright (C) by University of Illinois 2023
 */
#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define BENV_WITHMPI 1
#include "benvutil.h"

/* OPTIONS_N_INTS is the number of integers in the options structure to
   broadcast to the other processes.  Make sure to update this define
   when any changes are made to the structure. Similar for the doubles */
#define OPTIONS_N_INTS 11
#define OPTIONS_N_DOUBLES 4
typedef struct {
    int verbose;     /* Provide more information about the operation of code */
    int debug;       /* Use an artificial mesh topology for debugging */
    int progress;    /* If true, show progress while running tests */
    int printMap;    /* Print the mapping of processes to the topology */
    int isThreaded;  /* Set if use multiple threads for communication */
    int nMsgSizes;   /* Number of message sizes provided */
    int nProcNum;    /* Number of elements in nprocarray */
    int nTrials;     /* Repeat the test this many times */
    int ntestMIN;    /* ntest must be at least this size */
    int ntestMAX;    /* ntest much be no larger than this size */
    int ntestinfo;   /* if true, write ntestinfo to same output as outName */
    double ntestWT;  /* Multiple of Wtick to use in determining ntest */
    double ntestS;   /* Latency to use in determining ntest */
    double ntestR;   /* Inverse bandwidth to use in determining ntest */
    double ntestWtick; /* Use this value for wtick instead of MPI_Wtick */
    int *msgsizes;   /* Message sizes to use */
    int *nprocarray; /* Array of number of processes for tests */
    char *mapName;  /* Name for the mapping file. Only used on rank 0 */
    char *outName;  /* Name for output file.  If null, use stdout */
} options_t;

/* This type lets us save information about the computed values of ntest */
typedef struct ntestinfo_t {
    int    ntestval;      /* ntest value used */
    double acttime;       /* actual time for test */
} ntestinfo_t;

/* States for the communication call. */
enum cstate_t { IS_MASTER, IS_PARTNER, IS_IDLE };
typedef enum cstate_t commstate_t;

/* Communication test routine */
typedef double (commroutine_t)(MPI_Comm, commstate_t st, int mrank, int prank,
			       int msgsize, int ntest, int *sbuf, int *rbuf);

/* We use this to simplify two-dimensional indexing into an array */
#define IND2D(_i,_j,_ncol) ((_j)+(_i)*(_ncol))

#define MAX_HW_DEPTH 16

int getOptions(int argc, char *argv[], options_t *options);
int checkNodeDecomp(hwdesc_t *hw, int hwdepth);
int getPartners(hwdesc_t *hw, int hwdepth,
		int **masterranks, int **partnerranks, int *nranks);
int runTests(MPI_Comm comm, int nmsgsizes, const int msgsizes[],
	     int nranks, const int masterranks[], const int partnerranks[],
	     int nprocnum, const int nprocarray[],
	     commroutine_t *commroutine, double *timeArray,
	     ntestinfo_t *ntestinfo);
int examineTiming(MPI_Comm comm, double *timeArray, int nmsgsize, int nranks,
		  int nprocnum, int nTrials, double **tarray);
int printReport(FILE *, MPI_Comm comm, hwdesc_t hw[], int hwdepth,
		int nranks, int *masterranks, int *partnerranks,
		int nprocnum, int *nprocarray,
		int nmsgsizes, int *msgsizes, double *timeArray);

double commBlockingSend(MPI_Comm comm, commstate_t st, int mrank, int prank,
			int len, int ntest, int *sbuf, int *rbuf);
int ntestInit(options_t *options, MPI_Comm comm);
int ntestGetVal(int msgsize, int nc);
double ntestTimeEst(int len);
int ntestPrintInfo(FILE *fp, int nmsgs, int msgsizes[], ntestinfo_t ntestinfo[]);

/* For simplicity in debugging */
static int verbose = 0, dbg_wrank;
#define DBG(_s) do { if (verbose > 0 && dbg_wrank == 0) {\
	    fputs(_s,stderr);fflush(stderr);}} while(0)
#define DBG1(_s,_a) do { if (verbose > 0 && dbg_wrank == 0) {	\
	    fprintf(stderr,_s,_a);fflush(stderr);}} while(0)
#define DBG2(_s,_a,_b) do { if (verbose > 0 && dbg_wrank == 0) {	\
	    fprintf(stderr,_s,_a,_b);fflush(stderr);}} while(0)

/* If true (see getOptions --progress), wrank==0 gives indication of progress */
static int showProgress = 0;

/*
 * This is an updated version of mpptest.
 *
 * It adds support for measuring performance between all processes on a
 * node, communicating with their counterparts on another node.
 *
 * It does not include many of the adaptive features of mpptest
 */

int main(int argc, char **argv)
{
    hwdesc_t hw[MAX_HW_DEPTH];
    int      hwdepth, rc, required, provided, wsize, wrank;
    int      *masterranks, *partnerranks, nranks;
    options_t options;
    double   *timeArray = 0;
    int      tsize; /* Number of elements in a single block of timeArray */
    ntestinfo_t *ntestinfo=0;
    commroutine_t *commroutine = commBlockingSend;

    /* Get command line information, including message size range */
    if (getOptions(argc, argv, &options)) {
	/* We can't use MPI_Abort until we call MPI_Init_thread */
	return 1;
    }

    if (options.isThreaded)
	required = MPI_THREAD_MULTIPLE;
    else
	required = MPI_THREAD_SINGLE;
    MPI_Init_thread(&argc, &argv, required, &provided);
    MPI_Comm_rank(MPI_COMM_WORLD, &wrank);
    MPI_Comm_size(MPI_COMM_WORLD, &wsize);

    if (options.isThreaded && provided != MPI_THREAD_MULTIPLE) {
	if (wrank == 0)
	    fprintf(stderr,
	    "MPI_THREAD_MULTIPLE not provided and required for thread tests\n");
	MPI_Abort(MPI_COMM_WORLD, 1);
    }
    /* In most MPI implementations, all processes will have the options.
       But for best portability, rank 0 will broadcast to the others. Note
       that mapName and outName are only used on rank 0, and is not broadcast.
       The values in options are assumed to be 7 consecutively stored
       integers */
    MPI_Bcast(&options, OPTIONS_N_INTS, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&options.ntestWT, OPTIONS_N_DOUBLES,
	      MPI_DOUBLE, 0, MPI_COMM_WORLD);

    /* Get the msg sizes if defined as an array of values */
    if (options.nMsgSizes > 0)
	MPI_Bcast(options.msgsizes, options.nMsgSizes, MPI_INT,
		  0, MPI_COMM_WORLD);
    else {
	/* Create the array of values from 1 to 32*1024 by powers of 2 */
	options.msgsizes = BENV_GetSizesMult(1, 32*1024, 2.0,
					     &options.nMsgSizes);
    }

    /* nprocarray set after getPartners */

    ntestInit(&options, MPI_COMM_WORLD);
    ntestinfo = (ntestinfo_t *)malloc(options.nMsgSizes * sizeof(ntestinfo_t));

    /* Set debugging options */
    verbose   = options.verbose;
    dbg_wrank = wrank;
    if (options.progress && wrank == 0) showProgress = 1;

    DBG("Done with options\n");

    /* Get information on the nodes and processes, so that we can
       determine the communication partners */
    /* First set any debugging options */
    if (options.debug) {
	BENV_NodeCvarSet("ppn", wsize/2);
	BENV_NodeCvarSet("debug", 1);
    }
    rc = BENV_NodeGetComms(MPI_COMM_WORLD, hw, MAX_HW_DEPTH, &hwdepth);

    DBG1("Done with NodeGetComms (depth = %d)\n", hwdepth);

    /* Sanity check that we have a usable decomposition */
    rc = checkNodeDecomp(hw, hwdepth);
    if (rc) {
	MPI_Abort(MPI_COMM_WORLD, 1);
    }

    /* Determine the communication partners */
    rc = getPartners(hw, hwdepth, &masterranks, &partnerranks, &nranks);

    /* Check that nprocnum and nprocarray is consistent. If nprocnum
     was not set, set it here. */
    /* Get the number of processes to use for the tests */
    if (options.nProcNum > 0) {
	MPI_Bcast(options.nprocarray, options.nProcNum, MPI_INT,
		  0, MPI_COMM_WORLD);
	/* Validate values: in range and monotone increasing */
	if (wrank == 0) {
	    int prevval = -1;
	    for (int i=0; i<options.nProcNum; i++) {
		int curval = options.nprocarray[i];
		if (curval < 1 || curval > nranks) {
		    fprintf(stderr, "Invalid nprocs value[%d]=%d\n",
			    i, options.nprocarray[i]);
		    MPI_Abort(MPI_COMM_WORLD, 1);
		}
		if (curval <= prevval) {
		    fprintf(stderr, "Nonmonotone nprocs values %d <= %d\n",
			    curval, prevval);
		    MPI_Abort(MPI_COMM_WORLD, 1);
		}
		prevval = curval;
	    }
	}
    }
    else {
	/* Create the array of values from 1 to 32*1024 by powers of 2 */
	options.nprocarray = BENV_GetSizesArith(1, nranks, 1,
						&options.nProcNum);
    }


    DBG1("Done with getPartners (nranks = %d)\n", nranks);
    tsize     = options.nProcNum * options.nMsgSizes;
    timeArray = (double *)malloc(tsize*options.nTrials*sizeof(double));
    if (!timeArray) {
	fprintf(stderr, "Unable to allocate timeArray\n");
	fflush(stderr);
	MPI_Abort(MPI_COMM_WORLD, 0);
    }
    /* Perform the tests, saving the timing data.
       Includes option to run tests multiple times, not just as an average
       over runs at a give size.
       Pass the function used to perform the communication. The outer framework
       is the same. On the process with rank 0 in hw[0].comm, fills in
       timerArray (nranks x nMsgSizes array of doubles) with the time
       returned by commroutine.
    */
    for (int trial=0; trial<options.nTrials; trial++) {
	DBG1("Running tests for trial %d\n", trial);
	if (showProgress) fprintf(stdout, "Trial %d\n", trial);
	rc = runTests(hw[0].comm,
		      options.nMsgSizes, options.msgsizes,
		      nranks, masterranks, partnerranks,
		      options.nProcNum, options.nprocarray,
		      commroutine,
		      timeArray+trial*tsize, ntestinfo);
    }
    DBG("Tests completed\n");

    /* Generate report */
    {
    FILE *fp;
    double *tarray=0;
    if (wrank == 0) {
	if (options.outName) {
	    fp = fopen(options.outName, "w");
	    if (!fp) {
		fprintf(stderr, "Could not open %s for output\n",
			options.outName);
		MPI_Abort(MPI_COMM_WORLD, 1);
	    }
	}
	else fp = stdout;
    }
    if (options.printMap) {
	FILE *fm = fp;
	if (options.mapName) {
	    fm = fopen(options.mapName, "w");
	}
	BENV_HwdescPrintAll(fm, hw[0].comm, hw, hwdepth, 0);
	DBG("Output Hwdesc map\n");
	if (options.mapName) {
	    fclose(fm);
	}
    }

    examineTiming(hw[0].comm, timeArray, options.nMsgSizes, nranks,
		  options.nProcNum, options.nTrials, &tarray);
    if (wrank == 0) {
	printReport(fp, hw[0].comm,
		    hw, hwdepth,
		    nranks, masterranks, partnerranks,
		    options.nProcNum, options.nprocarray,
		    options.nMsgSizes, options.msgsizes,
		    tarray+tsize);  /* Use the minimum time */
	if (options.ntestinfo) {
	    fprintf(fp, "\nInformation on the selection of ntest used in these tests\n");
	    ntestPrintInfo(fp, options.nMsgSizes, options.msgsizes, ntestinfo);
	}
	if (options.outName) fclose(fp);
    }
    if (tarray) free(tarray);
    }
    /* Cleanup and free */
    free(timeArray);
    free(masterranks);
    free(partnerranks);

    MPI_Finalize();
    return 0;
}

/* ------------------------------------------------------------------------ */
static int argGetint(int i, int argc, char **argv, int *val);
static int argGetdouble(int i, int argc, char **argv, double *val);


int getOptions(int argc, char *argv[], options_t *options)
{
    int i;
    /* Set defaults */
    options->verbose    = 0;
    options->debug      = 0;
    options->progress   = 0;
    options->printMap   = 0;
    options->isThreaded = 0;
    options->nTrials    = 1;
    options->ntestMIN   = -1;
    options->ntestMAX   = -1;
    options->ntestinfo  = 0;
    options->ntestWT    = 0.0;
    options->ntestS     = 0.0;
    options->ntestR     = 0.0;
    options->ntestWtick = 0.0;
    options->mapName    = 0;
    options->outName    = 0;
    options->nMsgSizes  = 0;
    options->msgsizes   = 0;

    for (i=1; i<argc; i++) {
	//printf("Processing arg %s\n", argv[i]);
	if      (strcmp(argv[i],"-v") == 0)  {
	    /* This allows multiple -v options to set higher levels of
	       debugging */
	    options->verbose++;
	}
	else if (strcmp(argv[i], "-d") == 0) options->debug = 1;
	else if (strcmp(argv[i], "-progress") == 0) options->progress = 1;
	else if (strcmp(argv[i], "-t") == 0) options->isThreaded = 1;
	else if (strcmp(argv[i], "-o") == 0) {
	    i++;
	    if (i < argc)
		options->outName = strdup(argv[i]);
	    else {
		fprintf(stderr, "-o missing value\n");
		fflush(stderr);
		return 1;
	    }
	}
	else if (strcmp(argv[i],"-pm") == 0) options->printMap = 1;
	else if (strcmp(argv[i],"-no-pm") == 0) options->printMap = 0;
	else if (strcmp(argv[i],"-pname") == 0) {
	    i++;
	    if (i < argc)
		options->mapName = strdup(argv[i]);
	    else {
		fprintf(stderr, "-pname missing value\n");
		fflush(stderr);
		return 1;
	    }
	}
	else if (strcmp(argv[i], "-sizes") == 0) {
	    i++;
	    if (i < argc) {
		options->msgsizes = BENV_RstringToArray(argv[i],
							&options->nMsgSizes);
	    }
	    else {
		fprintf(stderr, "-sizes missing value\n");
		fflush(stderr);
		return 1;
	    }
	}
	else if (strcmp(argv[i], "-nprocs") == 0) {
	    i++;
	    if (i < argc) {
		options->nprocarray = BENV_RstringToArray(argv[i],
							  &options->nProcNum);
	    }
	    else {
		fprintf(stderr, "-nprocs missing value\n");
		fflush(stderr);
		return 1;
	    }
	}
	else if (strcmp(argv[i], "-ntrials") == 0) {
	    i++;
	    if (argGetint(i, argc, argv, &options->nTrials))
		return 1;
	}
	else if (strcmp(argv[i], "-ntest-info") == 0)
	    options->ntestinfo = 1;
	else if (strcmp(argv[i], "-ntest-min") == 0) {
	    i++;
	    if (argGetint(i, argc, argv, &options->ntestMIN))
		return 1;
	}
	else if (strcmp(argv[i], "-ntest-max") == 0) {
	    i++;
	    if (argGetint(i, argc, argv, &options->ntestMAX))
		return 1;
	}
	else if (strcmp(argv[i], "-ntest-wt") == 0) {
	    i++;
	    if (argGetdouble(i, argc, argv, &options->ntestWT))
		return 1;
	}
	else if (strcmp(argv[i], "-ntest-s") == 0) {
	    i++;
	    if (argGetdouble(i, argc, argv, &options->ntestS))
		return 1;
	}
	else if (strcmp(argv[i], "-ntest-r") == 0) {
	    i++;
	    if (argGetdouble(i, argc, argv, &options->ntestR))
		return 1;
	}
	else if (strcmp(argv[i], "-ntest-wtick") == 0) {
	    i++;
	    if (argGetdouble(i, argc, argv, &options->ntestWtick))
		return 1;
	}
	else {
	    fprintf(stderr, "Unrecognized option %s\n", argv[i]);
	    fflush(stderr);
	    return 1;
	}
    }
    return 0;
}

static int argGetint(int i, int argc, char **argv, int *val)
{
    if (i < argc) {
	/* FIXME: This should check for valid input */
	*val = atoi(argv[i]);
    }
    else {
	fprintf(stderr, "%s missing value\n", argv[i-1]);
	fflush(stderr);
	return 1;
    }
    return 0;
}

static int argGetdouble(int i, int argc, char **argv, double *val)
{
    if (i < argc) {
	/* FIXME: This should check for valid input */
	*val = strtod(argv[i],NULL);
    }
    else {
	fprintf(stderr, "%s missing value\n", argv[i-1]);
	fflush(stderr);
	return 1;
    }
    return 0;
}

/*
 * TODO: Add a way to run on a subset of nranks partners - note that
 * we still need the full set of partners, so that we can run with
 * that many partners.  But rather than, for example, running with 1,
 * 2, ... 128 concurrently communication processes, we'd like to have
 * the option to run with 1..16,18,20,..,32, 36, 40, .. 64, 80, 96,
 * .. 128 (eg, number of partners is 1:16,18:2:32,36:4:64,80:16:128)
 *
 * Input Parameters:
 */
int runTests(MPI_Comm comm, int nmsgsizes, const int msgsizes[],
	     int nranks, const int masterranks[], const int partnerranks[],
	     int nprocnum, const int nprocarray[],
	     commroutine_t *commroutine, double *timeArray,
	     ntestinfo_t *ntestinfo)
{
    commstate_t state;
    int lrank, i, len, mrank, prank, nr, maxsize, ntest;
    int *sbuf, *rbuf;
    double t;
    int curprocnumidx;

    MPI_Comm_rank(comm, &lrank);

    maxsize = msgsizes[nmsgsizes-1];
    sbuf = (int *)malloc(maxsize * sizeof(int));
    rbuf = (int *)malloc(maxsize * sizeof(int));
    for (i=0; i<maxsize; i++) {
	sbuf[i] = i;
	rbuf[i] = -i;
    }

    if (showProgress) {
	fputs("\tsize:", stdout);
	fflush(stdout);
    }
    for (i=0; i<nmsgsizes; i++) {
	len   = msgsizes[i];
	if (showProgress) {
	    fprintf(stdout,"%d,", len);
	    fflush(stdout);
	}
	/* Move ntestGetVal into nr loop if ntestGetVal used nc */
	ntest = ntestGetVal(len, 1);
	state = IS_IDLE;
	curprocnumidx = 0;
	DBG2("About to run test %d with ntest=%d\n", i, ntest);
	for (nr=0; nr<nranks; nr++) {
	    /* Once a process becomes active, it stays active for this
	       message size */
	    if (state == IS_IDLE) {
		/* are we a master or partner or still idle? */
		if (lrank == masterranks[nr]) {
		    state = IS_MASTER;
		    mrank = lrank;
		    prank = partnerranks[nr];
		}
		else if (lrank == partnerranks[nr]) {
		    state = IS_PARTNER;
		    mrank = masterranks[nr];
		    prank = lrank;
		}
	    }

	    /* Only run the tests if the number of ranks is the current
	       value in nprocarray */
	    if (nr + 1 == nprocarray[curprocnumidx]) {
		if (nr == 0) t = MPI_Wtime();
		timeArray[IND2D(i,curprocnumidx,nprocnum)] =
		    (*commroutine)(comm, state, mrank, prank, len, ntest,
				   sbuf, rbuf);
		/* For the first test (2 processes), record the ntest and
		   the actual time for the test */
		if (nr == 0) {
		    ntestinfo[i].acttime  = MPI_Wtime() - t;
		    ntestinfo[i].ntestval = ntest;
		}
		curprocnumidx++;
	    }

	}
	DBG1("Done running test %d\n",i);
    }
    if (showProgress) {
	fputc('\n', stdout);
	fflush(stdout);
    }
    free(rbuf);
    free(sbuf);
    return 0;
}

/*
 * Sanity check on the decomposition. Return 0 on success, non-zero on
 * failure. Rank 0 in the hw[0].comm communicator prints messages to
 * stderr.
 */
int checkNodeDecomp(hwdesc_t *hw, int hwdepth)
{
    int wrank, nsize, psize;

    MPI_Comm_rank(hw[0].comm, &wrank);
    /* Things to check:
       2 distinct comms at level 1 (level 0 is MPI_COMM_WORLD)
       Both comms have the same size.
       ToDo: if depth > 2 and we take into account the lower levels,
       we will need to check them as well
    */
    if (hwdepth < 2) {
	if (wrank == 0) {
	    fprintf(stderr, "hwdepth = %d; need at least 2\n", hwdepth);
	    fflush(stderr);
	}
	return 1;
    }
    if (hw[1].nDistinct != 2) {
	if (wrank == 0) {
	    fprintf(stderr, "Expected 2 nodes, found %d %s\n",
		    hw[1].nDistinct, hw[1].descstr);
	    fflush(stderr);
	}
	return 1;
    }
    MPI_Comm_size(hw[1].comm, &nsize);
    MPI_Comm_size(hw[0].comm, &psize);
    if (nsize*2 != psize) {
	if (wrank == 0) {
	    fprintf(stderr, "Expected nodes to have same size; node on 0 size %d\n",
		    nsize);
	    fflush(stderr);
	}
	return 1;
    }

    return 0;
}
/*
 * Get the ranks in hw[0].comm of the master (node 0) and partner (node 1)
 * processes. Assumes a correct decomposition (see checkNodeDecomp)
 */
int getPartners(hwdesc_t *hw, int hwdepth,
		int **masterranks, int **partnerranks, int *nranks)
{
    int       nrank, nsize, *mranks, *pranks, lrank;

    /* The leader processes in each comm will determine the ranks in
       hw[0].comm of the processes in the respective comm. These are then
       broadcast to *all* processes in hw[0].comm. */
    MPI_Comm_rank(hw[1].comm, &nrank);
    MPI_Comm_size(hw[1].comm, &nsize);
    mranks = (int *)malloc(nsize*sizeof(int));
    pranks = (int *)malloc(nsize*sizeof(int));
    if (!mranks || !pranks) {
	fprintf(stderr, "Unable to allocated %d words for ranks\n", 2*nsize);
	fflush(stderr);
	MPI_Abort(hw[0].comm, 1);
    }
    /* Are we a locak leader? */
    if (nrank == 0) {
	MPI_Group pgroup, ngroup, *ranks;
	int *allranks;
	allranks = (int *)malloc(nsize*sizeof(int));
	if (!allranks) {
	    fprintf(stderr, "Unable to allocated %d words for ranks\n", nsize);
	    fflush(stderr);
	    MPI_Abort(hw[0].comm, 1);
	}
	if (hw[1].idx == 0) ranks = mranks;
	else                ranks = pranks;
	for (int i=0; i<nsize; i++) allranks[i] = i;
	MPI_Comm_group(hw[0].comm, &pgroup);
	MPI_Comm_group(hw[1].comm, &ngroup);
	MPI_Group_translate_ranks(ngroup, nsize, allranks, pgroup, ranks);
	MPI_Group_free(&ngroup);
	MPI_Group_free(&pgroup);
	free(allranks);
    }
    /* Broadcast the ranks */
    MPI_Bcast(mranks, nsize, MPI_INT, hw[1].leaders[0], hw[0].comm);
    MPI_Bcast(pranks, nsize, MPI_INT, hw[1].leaders[1], hw[0].comm);

    *masterranks  = mranks;
    *partnerranks = pranks;
    *nranks       = nsize;

    return 0;
}

int examineTiming(MPI_Comm comm, double *timeArray, int nmsgsizes, int nranks,
		  int nprocnum, int nTrials, double **tarray)
{
    int    i, trial, tsize;
    double *tmin, *tmax, *tavg;
    double t, tminval, tmaxval, tavgval;
    *tarray = 0;

    /* Determine the time values to use, by computing the min, max,
       and average over the trials */
    tsize = nmsgsizes * nprocnum;
    tavg = (double *)malloc(3*tsize*sizeof(double));
    tmin = tavg + tsize;
    tmax = tmin + tsize;

    /* Compute the values for the current process, then look over
       all process. Note that in Runtests, the time is taken as the
       max across all communicating processes */

    /* We could flip this around so all accesses are stride 1 */
    for (i=0; i<tsize; i++) {
	double t;
	t = timeArray[i];
	tminval = tmaxval = tavgval = t;
	for (trial=1; trial<nTrials; trial++) {
	    t = timeArray[trial*tsize+i];
	    if (tminval > t) tminval = t;
	    if (tmaxval < t) tmaxval = t;
	    tavgval += t;
	}
	tmin[i] = tminval;
	tmax[i] = tmaxval;
	tavg[i] = tavgval / nTrials;
    }

    /* Note that the "other" values (min, max) are available at tarray + tsize
       and tarray + 2*tsize */
    *tarray = tavg;

    return 0;
}

/* Print the timing report. Only the process performing the output should
   call
*/
int printReport(FILE *fp, MPI_Comm comm, hwdesc_t hw[], int hwdepth,
		int nranks, int *masterranks, int *partnerranks,
		int nprocnum, int *nprocarray,
		int nmsgsizes, int *msgsizes, double *timeArray)
{
    int    i, nr;
    /* Print the column headings */
    fprintf(fp, "\nInts");
    for (int nr=0; nr<nprocnum; nr++)
	fprintf(fp, "\t%-8d", nprocarray[nr]);
    fprintf(fp, "\n");

    /* Print the timing data */
    for (i=0; i<nmsgsizes; i++) {
	fprintf(fp, "%d\t", msgsizes[i]);
	for (nr=0; nr<nprocnum; nr++) {
	    fprintf(fp, "%.2e\t", timeArray[IND2D(i,nr,nprocnum)]);
	}
	fprintf(fp, "\n");
    }
    fflush(fp);
    return 0;
}

/* Communication timing routines
 */
double commBlockingSend(MPI_Comm comm, commstate_t st, int mrank, int prank,
			int len, int ntest, int *sbuf, int *rbuf)
{
    double t = -1.0;
    int k, tag = 1 + /*nr * */ntest;

    MPI_Barrier(comm);
    if (st == IS_MASTER) {
	t = MPI_Wtime();
	for (k=0; k<ntest; k++) {
	    MPI_Send(sbuf,len,MPI_INT,prank,tag+k,comm);
	    MPI_Recv(rbuf,len,MPI_INT,prank,tag+k,comm, MPI_STATUS_IGNORE);
	}
	t = (MPI_Wtime() - t)/ntest;
    }
    else if (st == IS_PARTNER) {
	t = MPI_Wtime();
	for (k=0; k<ntest; k++) {
	    MPI_Recv(rbuf,len,MPI_INT,mrank,tag+k, comm, MPI_STATUS_IGNORE);
	    MPI_Send(sbuf,len,MPI_INT,mrank,tag+k, comm);
	}
	t = (MPI_Wtime() - t)/ntest;
    }
    /* else state is IS_IDLE; ignore */
    MPI_Allreduce(MPI_IN_PLACE, &t, 1, MPI_DOUBLE, MPI_MAX, comm);
    return t;
}

/* ----------------------------------------------------------------------- */
/* These routines determine the number of tests to perform. The goal is
   to ensure that the the tests take at least ntestWT times the resolution
   of the timer, which is given (we hope!) by MPI_Wtick. This is estimated
   by assuming the simple performance model, T = s = rn, where s is the latency
   and r is the inverse rate. Provisions are made to extend this to the maxrate
   model, T = s + (1/max(1/rmax,nc*1/r1))n, where nc is the number of
   concurrently communicating processes. Using this formula, ntest can
   be estimate as

   ntest * 2(s+rn) = ntestWT * MPI_Wtick

   (2 since the test is a ping-pong, so there are two communiations in each
   test).

   Thus,

       ntest = ntestWT * MPI_Wtick / (2 (s + rn))

   As a sanity check, ntest is also constrained to be between ntestMIN and
   ntestMAX.
*/
static double ntestWT, /* Multiple of Wtick to use in determining ntest */
    ntestS,            /* Latency to use in determining ntest */
    ntestR,            /* Inverse bandwidth to use in determining ntest */
    ntestWtick;        /* Use this value for wtick instead of MPI_Wtick */
static int ntestMIN,
    ntestMAX;          /* These provide a valid range for ntest values */

int ntestInit(options_t *options, MPI_Comm comm)
{
    double wtick;

    /* Save the values from options, setting defaults if necessary */
    if (options->ntestMIN <= 0)
	ntestMIN = 10;
    else
	ntestMIN = options->ntestMIN;

    if (options->ntestMAX <= 0)
	ntestMAX = 10000;
    else
	ntestMAX = options->ntestMAX;

    if (options->ntestWT <= 0)
	ntestWT = 100;
    else
	ntestWT = options->ntestWT;

    if (options->ntestS <= 0)
	ntestS = 2.0e-6;
    else
	ntestS = options->ntestS;

    if (options->ntestR <= 0)
	/* Default is 1 gigaword/sec, or an inverse of 1e-9 */
	ntestR = 1.0e-9;
    else
	ntestR = options->ntestR;

    if (options->ntestWtick <= 0) {
	double wt = MPI_Wtick();
	/* Also examine the average cost of a call to
	   MPI_Wtime - which is also relevant to the accuracy of the
	   times from MPI_Wtime */
	double t, t1;
	t = MPI_Wtime();
	for (int i=0; i<20; i++) {
	    t1 = MPI_Wtime();
	}
	t = (t1 - t) / 20;
	if (wt < t) wt = t;

	MPI_Allreduce(MPI_IN_PLACE, &wt, 1, MPI_DOUBLE, MPI_MAX, comm);
	if  (wt > 1.0e-9)
	    ntestWtick = wt;
	else
	    /* FIXME: if wt is 0, MPI_Wtick is erroneous */
	    ntestWtick = 1.0e-9;
    }
    else
	ntestWtick = options->ntestWtick;

    //fprintf(stderr, "s = %.2e, r = %.2e, wt = %.2e, wtick = %.2e\n",
    //	    ntestS, ntestR, ntestWT , ntestWtick);
    //fflush(stderr);
    return 0;
}

int ntestGetVal(int msgsize, int nc)
{
    int ntest;

    ntest = ntestWT * ntestWtick / ( 2.0 * (ntestS + ntestR * msgsize));

    if (ntest < ntestMIN) ntest = ntestMIN;
    if (ntest > ntestMAX) ntest = ntestMAX;

    return ntest;
}

/* This gives the time that was estimated that each test would take */
double ntestTimeEst(int len)
{
    return 2.0 * (ntestS + ntestR * len);
}

int ntestPrintInfo(FILE *fp, int nmsgs, int msgsizes[], ntestinfo_t ntestinfo[])
{
    fprintf(fp, "Model ntest = %.2e * %.2e / (2 * (%.2e + %.2e * len))\n",
	    ntestWT, ntestWtick, ntestS, ntestR);
    fprintf(fp, "Ints\tntest\test time\tact time\n");
    for (int i=0; i<nmsgs; i++) {
	fprintf(fp, "%d\t%d\t%.2e\t%.2e\n", msgsizes[i], ntestinfo[i].ntestval,
		ntestTimeEst(msgsizes[i])*ntestinfo[i].ntestval,
		ntestinfo[i].acttime);
    }
    fflush(fp);
    return 0;
}
